//--------------------------------------
// Forest and Tree objects
//--------------------------------------
// Copyright (C) 1993 Nicholas Centanni
//--------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <time.h>
#include <graphics.h>
#include <alloc.h>
#include <string.h>
#include <opus.h>
#include <op_graph.h>
#include <op_view.h>
#include <op_color.h>
#include "nb_app.h"
#include "ob_tree.h"

#ifdef TREES
//-------------------------------------------------------------------------
// Tree constructor
//-------------------------------------------------------------------------
Tree::Tree(Environ *env)
{
  // Assign connection to the service classes
  viewer = env->viewer;
  graphics = env->graphics;
  memset(vertices, 0, sizeof(vertices));
  memset(lines, 0, sizeof(lines));

  // Determine a random height
  height = get_random(TREE_MIN_HT, TREE_MAX_HT);
  base = &vertices[0].world.z;

  // Clear flags
  tree_z = 0L;
  det_max = fading_out = fading = OFF;
}

//-------------------------------------------------------------------------
// Translate all tree vertices
//-------------------------------------------------------------------------
int Tree::translate()
{
  int page = graphics->active_page;

  // Check the trunk first to determine tree visibility
  int v1 = viewer->translate(&vertices[0].world, &vertices[0].pixel[page], &tree_z);
  int v2 = viewer->translate(&vertices[1].world, &vertices[1].pixel[page]);
  if (!v1 && !v2)
    return 0;

  for (int i = 1; i < LINE_TOT; i++)
  {
    if (lines[i].active)
    {
      viewer->translate(&vertices[lines[i].st].world,
                        &vertices[lines[i].st].pixel[page]);
      viewer->translate(&vertices[lines[i].en].world,
                        &vertices[lines[i].en].pixel[page]);
    }
  }
  return 1;
}

//-------------------------------------------------------------------------
// Draw the tree
//-------------------------------------------------------------------------
void Tree::draw()
{
  int page = graphics->active_page;
  int color = EARTH_BROWN;
  tr_line *ln = lines;

  // Draw all visible tree branches
  for (int i = 0; i < LINE_TOT; i++, ln++)
  {
    if (ln->active)
    {
      i_screen *p1 = &vertices[ln->st].pixel[page];
      i_screen *p2 = &vertices[ln->en].pixel[page];
      if (p1->visible && p2->visible)
        n_line(p1->x, p1->y, p2->x, p2->y, color, 0);
    }
    color = EARTH_MED_GREEN;
  }
}

//-------------------------------------------------------------------------
// Erase the tree
//-------------------------------------------------------------------------
void Tree::erase()
{
  int page = graphics->active_page;
  tr_line *ln = lines;

  // Draw all visible tree branches
  for (int i = 0; i < LINE_TOT; i++, ln++)
  {
    if (ln->active || fading_out)
    {
      i_screen *p1 = &vertices[ln->st].pixel[page];
      i_screen *p2 = &vertices[ln->en].pixel[page];
      if (p1->visible && p2->visible)
        n_line(p1->x, p1->y, p2->x, p2->y, EARTH_BLACK, 0);
    }
  }
}

//------------------------------------------------------------------------
// Flag a tree's inactive branches
//------------------------------------------------------------------------
void Tree::detail(int state)
{
  int i;

  if (state == ON)
  {
    for (i = 0; i < LINE_TOT; i++)
      lines[i].active = 1;
    det_max = ON;
  }
  else
  {
    for (i = 0; i < LINE_TOT; i += 2)
    {
      lines[i].active = 1;
      lines[i+1].active = 0;
    }
    det_max = OFF;
  }
}

//------------------------------------------------------------------------
// Build a random tree at the given coordinates
//------------------------------------------------------------------------
void Tree::build(long x, long y, long z)
{
  int v_index = 0;
  tr_vertex *base = &vertices[0];

  // Define the trunk as PT-0 to PT-1
  vertices[v_index].world.x = x;
  vertices[v_index].world.y = y;
  vertices[v_index].world.z = z;
  lines[0].st = v_index++;
  vertices[v_index].world.x = x;
  vertices[v_index].world.y = y + height;
  vertices[v_index].world.z = z;
  lines[0].en = v_index++;

  // Define the branches
  double branch_start = 0.45;        // Branch starting height
  double length_reducer = 1.8;       // Length reduction limiter
  double elev = (double)height * branch_start;
  double min_length = height * 0.02;
  double length = height * 0.20;
  double ht_increment = (height - (height * 0.03) - elev) / ((LINE_TOT-1) / 3);
  double height_diff = 0.20;         // Height increment randomness margin
  int height_margin = ht_increment * height_diff;
  int heading_increment = 1200;
  double heading_diff = 0.20;        // Heading randomness margin
  int heading_margin = (double)heading_increment * heading_diff;
  int branch_pitch = 3100;
  double pitch_diff = 0.025;         // Pitch randomness margin
  int pitch_margin = branch_pitch * pitch_diff;

  int ht_cnt = 0;
  int heading = random(3600);
  for (int i = 1; i < LINE_TOT; i++)
  {
    // Determine the branch START point
    vertices[v_index].world.x = base->world.x;
    vertices[v_index].world.y = base->world.y + elev + ht_increment;
    vertices[v_index].world.z = base->world.z;

    // Determine the branch END point
    int pitch = branch_pitch + get_random(-pitch_margin, pitch_margin);
    double n_length = min_length + 
      (length * (1.0 - (elev/height)) * length_reducer);

    // Calculate the branch end point
    l_point start, end;
    viewer->move(&vertices[v_index].world, &vertices[v_index+1].world,
                  heading, pitch, (long)n_length);

    // Record the endpoints
    lines[i].st = v_index++;
    lines[i].en = v_index++;

    // Increment settings for next branch
    if (++ht_cnt == 3)
    {
      elev += ht_increment + get_random(-height_margin, height_margin);
      if (elev > height)
       elev = height - height_margin;
      ht_cnt = 0;
      heading = random(3600);
    }
    else
    {
      heading += heading_increment + get_random(-heading_margin, heading_margin);
      viewer->clip_angle(&heading);
    }
  }

  // Set tree to minimum detail
  detail(OFF);
}

//---------------------------------------------------------------------------
// Forest constructor
//---------------------------------------------------------------------------
Forest::Forest(Environ *e, int t_max, long x_limit, long z_limit)
{
  // Copy the supplied values
  env = e;
  graphics = env->graphics;
  max_trees = t_max;
  memset(visible, 0, sizeof(visible));
  tree_tot = active_trees = 0;
  x_lim = x_limit;
  z_lim = z_limit;
  z_margin = z_limit * 0.05;

  // Create the first crop of trees
  tree = 0L;
  create_trees();
}

//---------------------------------------------------------------------------
// Forest destructor
//---------------------------------------------------------------------------
Forest::~Forest()
{
  delete_trees();
}

//---------------------------------------------------------------------------
// Remove the existing group of trees
//---------------------------------------------------------------------------
void Forest::delete_trees()
{
  if (tree)
  {
    for (int i = 0; i < tree_tot; i++)
      delete tree[i];
    delete tree;
  }
}

//---------------------------------------------------------------------------
// Build a group of trees
//---------------------------------------------------------------------------
void Forest::create_trees()
{
  // First remove trees if they already exist
  delete_trees();

  // Create an array of Tree class pointers
  if (!(tree = new Tree*[max_trees]))
    bail("Mem:tree pointers");

  // Add trees to the scene
  #ifdef BENCHMARK
  srand(1);
  #else
  randomize();
  #endif
  for (int i = 0; i < max_trees; i++)
  {
    if ((tree[i] = new Tree(env)) != NULLP)
    {
      long x = -TREE_LINE + (random(TREE_LINE) * 2);
      long z = z_lim - (random(z_lim) * 2);
      tree[i]->build(x, 0, z);
      tree_tot++;
    }
    else
    {
      char msg[80];
      sprintf(msg, "Mem:adding tree %d", i);
      bail(msg);
    }
  }
}

//---------------------------------------------------------------------------
// Increase the active trees total.
//---------------------------------------------------------------------------
void Forest::add_tree()
{
  if (active_trees < tree_tot)
    active_trees++;
}

//---------------------------------------------------------------------------
// Translate all active trees
//---------------------------------------------------------------------------
void Forest::translate()
{
  for (int i = 0; i < active_trees; i++)
    visible[i] = tree[i]->translate();
}

//---------------------------------------------------------------------------
// Display all trees
//---------------------------------------------------------------------------
void Forest::draw()
{
  for (int i = 0; i < active_trees; i++)
  {
    if (visible[i])
      tree[i]->draw();
  }
}

#ifdef BIRD
//---------------------------------------------------------------------------
// Display all trees that appear in BACK of the bird
//---------------------------------------------------------------------------
void Forest::draw_back()
{
  for (int i = 0; i < active_trees; i++)
  {
    if (visible[i] && (tree[i]->tree_z > bird->curr_z))
      tree[i]->draw();
  }
}

//---------------------------------------------------------------------------
// Display all trees that appear in FRONT of the bird
//---------------------------------------------------------------------------
void Forest::draw_front()
{
  for (int i = 0; i < active_trees; i++)
  {
    if (visible[i] && (tree[i]->tree_z <= bird->curr_z))
      tree[i]->draw();
  }
}
#endif

//---------------------------------------------------------------------------
// Erase all visible trees
//---------------------------------------------------------------------------
void Forest::erase()
{
  for (int i = 0; i < active_trees; i++)
    tree[i]->erase();
}

//---------------------------------------------------------------------------
// Move the trees
//---------------------------------------------------------------------------
void Forest::move(long distance)
{
  for (int i = 0; i < active_trees; i++)
  { 
    Tree *t = tree[i];

    if (*t->base > z_lim)
    {
      //
      // Move the tree to the other end of the scene
      //
      long x_new = -TREE_LINE + (random(TREE_LINE) * 2);
      long x_offset = x_new - t->vertices[0].world.x;
      long z_new = -z_lim + random(z_margin);
      long z_offset = z_new - t->vertices[0].world.z;
      for (int j = 0; j < VERT_TOT; j++)
      {
        t->vertices[j].world.z += z_offset;
        t->vertices[j].world.x += x_offset;
      }
    }
    else
    {
      // 
      // Move the tree forward
      //
      for (int j = 0; j < VERT_TOT; j++)
        t->vertices[j].world.z += distance;

      //
      // Handle changes in tree detail
      //
      if (*t->base > -TREE_FADE)
      { 
        if (*t->base < TREE_FADE)
        {
          if (!t->det_max)
            t->detail(ON);
        }
        else if (t->det_max)
        {
          t->detail(OFF);
          t->fading_out = ON;
          t->fading = 2;
        }
        else if (t->fading_out)
          if (--(t->fading) == 0)
            t->fading_out = OFF;
      }
    }
  }
}
#endif
